#include "defines.h"

(* Code emitter for IA32.
 * Instructions are taken from v8/src/assembler-ia32.c
 *
 * Good x86 source:
 * http://www.cs.uaf.edu/2005/fall/cs301/lecture/2005_12_07_x86_machinecode.html
 *)

open Utils;;

exception OperandError of string;;

(* Emitter type.
 * Should be created with 'make' and components accessed with the
 * getters and setters.
 *)
type 'a t =
    { pc     : int
    ; marks  : ('a * int) list
    ; buffer : string ref
    ; msize  : int
    ; delays : (int * ('a t -> 'a t)) list
    }
;;

(* internal types for the operand components *)
type sib   = int * int * int;;
type disp  = int;;
type modrm = int * int * sib option * disp option;;

(* type of targets *)
type target =
  | Reg of int
  | Int of int
  | Opr of modrm
;;

(* scale factors *)
type scale =
  | Scale1
  | Scale2
  | Scale4
  | Scale8
;;

(* conditions *)
type condition =
  | NoCondition
  | Overflow
  | NoOverflow
  | Below
  | AboveEqual
  | Equal
  | NotEqual
  | BelowEqual
  | Above
  | Sign
  | NotSign
  | ParityEven
  | ParityOdd
  | Less
  | GreaterEqual
  | LessEqual
  | Greater
  | Zero
  | NotZero
  | Negative
  | Positive
;;

(* register constants (values do not follow alphabetical) *)
let eax = Reg 0x00;;
let ecx = Reg 0x01;;
let edx = Reg 0x02;;
let ebx = Reg 0x03;;
let esp = Reg 0x04;;
let ebp = Reg 0x05;;
let esi = Reg 0x06;;
let edi = Reg 0x07;;

(* -------------------------------------------------- *)
(* Internal Helpers                                   *)
(* -------------------------------------------------- *)

let scale_value = function
  | Scale1 -> 0
  | Scale2 -> 1
  | Scale4 -> 2
  | Scale8 -> 3
;;

let condition_value = function
  | NoCondition        -> -1
  | Overflow           ->  0
  | NoOverflow         ->  1
  | Below              ->  2
  | AboveEqual         ->  3
  | Equal    | Zero    ->  4
  | NotEqual | NotZero ->  5
  | BelowEqual         ->  6
  | Above              ->  7
  | Sign    | Negative ->  8
  | NotSign | Positive ->  9
  | ParityEven         -> 10
  | ParityOdd          -> 11
  | Less               -> 12
  | GreaterEqual       -> 13
  | LessEqual          -> 14
  | Greater            -> 15
;;

let is_int8 i = i >= -128 && i <= 127;;

let operand_string = function
  | Int _ -> "integer"
  | Reg _ -> "register"
  | Opr _ -> "location"
;;

let operand_error s msg =
  raise (OperandError
	   ("operand error in function '" ^ s ^ ": " ^ msg))
;;

let operand_error1 s opr =
  raise (OperandError
	   ("unsupported operand type '"
	    ^ operand_string opr ^ "' in '" ^ s ^ "'"))
;;

let operand_error2 s opr1 opr2 =
  raise (OperandError
	   ("unsupported operand types '"
	    ^ operand_string opr1 ^ "' and '"
	    ^ operand_string opr2 ^ "' in '" ^ s ^ "'"))
;;

(* increases the size of the internal buffer by half of its current size *)
let grow e =
  let len  = String.length !(e.buffer) in
  let nlen = len + len / 2 in
  let buf  = String.create nlen in
    String.blit !(e.buffer) 0 buf 0 len;
    e.buffer := buf
;;

(* -------------------------------------------------- *)
(* Emitter Manipulation                               *)
(* -------------------------------------------------- *)

(* Emitting syntax:
 *
 * let e = Emitter.make 10
 *   $ mov x eax
 *   $ push eax
 * in contents e
 *
 * Will create an emitter with an initial internal buffer of 10
 * bytes. It will output a mov and push instruction and then return
 * a copy of the emitted bytes.
 *)
let ($) (x : 'a t) (y : 'a t -> 'a t) = y x;;

(* Create a new emitter with an internal buffer of size 's' *)
let make s : 'a t =
  { pc = 0
  ; marks = []
  ; buffer = ref (String.create s)
  ; msize = 0
  ; delays = []
  }
;;

(* Reset the emitter *)
let reset e =
  { pc = 0
  ; marks = []
  ; buffer = e.buffer
  ; msize = 0
  ; delays = []
  }
;;

(* Get the pc and marks of the emitter *)
let pc    e = e.pc;;
let marks e = e.marks;;

(* Get the size in bytes of the emitted code *)
let size e = max e.pc e.msize;;

(* Get a copy of the emitted code.
 * This creates a copy of the internal bufferstring that is as large
 * as the amount of code emitted since the last reset or the emitter
 * creation.
 *)
let contents e =
  String.sub !(e.buffer) 0 (size e)
;;

(* Update the pc of the emitter *)
let set_pc pc e =
  (if pc > size e then internal_error "attempt to set pc out of emitted code range");
  (if pc < 0 then internal_error "attempt to set a negative pc");
  { e with pc = pc; msize = size e }
;;

(* Add a mark at the current pc *)
let mark m e = 
  { e with marks = (m, e.pc) :: e.marks }
;;

(* Push an emitter construct, depending on some later information,
 * onto the delay stack. 
 * The construct will be run once at the time of the delay construct
 * with both parameters being the current emitter. When forced it will
 * again be run with new current emitter as the first argument and the
 * emitter at the time of delay as the second. If the output size of
 * the two runs does not match an exception it raised.
 *
 * Usage:
 * e $ op1
 *   $ op2
 *   $ delay (fun late -> op3)
 *   $ op4
 *   $ force
 *)
let delay f e =
  let o = f e e in
    { o with delays = (o.pc, fun n -> f n e) :: e.delays }
;;

(* Force a delayed construct from the delay stack.
 * The resulting emitter will be the same as the input except for the
 * latest delay that will have been poped and the side effects the
 * delayed construct performs on the byte buffer.
 * I.e. the pc will remain untouched.
 *)
let force e =
  match e.delays with
    | [] -> internal_error "no delayed emission to force"
    | (p, f) :: ds ->
	if (f e).pc != p
	then internal_error "forced bytes does not match precomputed";
	{ e with delays = ds }
;;

(* Pretty print a string of bytes in hex. *)
let print_buffer ff buf =
  for i=0 to String.length buf - 1 do
    fprintf ff "%2X@;" (Char.code buf.[i])
  done
;;

(* -------------------------------------------------- *)
(* Emitting Primitives                                *)
(* -------------------------------------------------- *)

let emit op e =
  let p = e.pc in
    if p >= String.length !(e.buffer) then grow e;
    (* set_pc checks for p < 0 so we may use unsafe set *)
    String.unsafe_set !(e.buffer) p (Char.chr op);
    { e with pc = p + 1 }
;;

let emit_int i e =
  e $ emit (i land 0xFF)
    $ emit ((i lsr  8) land 0xFF)
    $ emit ((i lsr 16) land 0xFF)
    $ emit ((i lsr 24) land 0xFF)
;;

let emit_operand ro o e = match o with
  | Opr (md, rm, sib, disp) ->
      e $ emit (((md lsl 6) lor (ro lsl 3)) lor rm)
	$ (match sib with
	     | None -> id
	     | Some (scale, index, base) ->
		 emit (((scale lsl 6) lor (index lsl 3)) lor base))
	$ (match disp, md with
	     | None, _ -> id
	     | Some d, 1 -> emit d	(*  8-bit *)
	     | Some d, 2 -> emit_int d	(* 32-bit *)
	     | _ -> operand_error "emit_operand" "invalid displacement or mod")
  | _ -> operand_error1 "emit_operand" o
;;

(* -------------------------------------------------- *)
(* Operand Constructors                               *)
(* -------------------------------------------------- *)

let operand_reg = function
  | Reg r -> Opr (3, r, None, None)
  | o -> operand_error1 "operand_reg" o
;;

let operand_mem = function
  | Reg 4 -> Opr (0, 4, Some (0, 4, 4), None) (* weird special case for esp *)
  | Reg r -> Opr (0, r, None, None)
  | o -> operand_error1 "operand_mem" o
;;

let operand_sib b i s = match b, i, scale_value s with
  | Reg base, Reg index, scale ->
      ASSERT(index != 4);  (* esp is not allowed as an index *)
      Opr (0, 4(* esp *), Some (scale, index, base), None)
  | _ -> operand_error2 "operand_sib" b i
;;

let operand_disp r d =
  let mksib = function
    | 4 -> Some (0, 4, 4) (* weird special case for esp *)
    | _ -> None
  in match r, d with
    | Reg r, 0 -> Opr (0, r, mksib r, None)                        (* nodisp *)
    | Reg r, disp when is_int8 d -> Opr (1, r, mksib r, Some disp) (*  8-bit *)
    | Reg r, disp -> Opr (2, r, mksib r, Some disp)                (* 32-bit *)
    | _ -> operand_error1 "operand_disp" r
;;

(* -------------------------------------------------- *)
(* IA32 Instructions                                  *)
(* -------------------------------------------------- *)

let nop e =
  e $ emit 0x90
;;

let ret r e = match r with
  | 0 -> e $ emit 0xC3
  | i ->
      e $ emit 0xC2
	$ emit (i land 0xFF)
	$ emit ((i lsr 8) land 0xFF)
;;

let mov s d e = match s, d with
  | Int s, Reg d ->
      e $ emit (0xB8 lor d)
	$ emit_int s
  | Int s, Opr _ ->
      e $ emit 0xC7
	$ emit_operand 0 d
	$ emit_int s
  | Reg s, Reg d ->
      e $ emit 0x8B
	$ emit_operand d (operand_reg (Reg s))
  | Opr _, Reg d ->
      e $ emit 0x8B
	$ emit_operand d s
  | Reg s, Opr _ ->
      e $ emit 0x89
	$ emit_operand s d
  | _ -> operand_error2 "mov" s d
;;

let pop d e = match d with
  | Reg d -> e $ emit (0x58 lor d)
  | _ -> operand_error1 "pop" d
;;

let push s e = match s with
  | Reg s -> e $ emit (0x50 lor s)
  | Int i ->
      e $ emit 0x68
	$ emit_int i
  | _ -> operand_error1 "push" s
;;

let call d e = match d with
  | Reg t ->
      e $ emit 0xFF
	$ emit_operand 2 (operand_reg (Reg t)) (* the opcode says /2 *)
  | Opr t ->
      e $ emit 0xFF
	$ emit_operand 2 (Opr t)
  | _ -> operand_error1 "call" d
;;

let add s d e = match s, d with
  | Int s, Reg 0 ->
      e $ emit 0x05
	$ emit_int s (* eax *)
  | Int s, Reg d ->
      e $ emit 0x81
	$ emit_operand 0 (operand_reg (Reg d))
	$ emit_int s
  | _ -> operand_error2 "add" s d
;;

let sub s d e = match s, d with
  | Int s, Reg 0 ->
      e $ emit 0x2D
	$ emit_int s (* eax *)
  | Int s, Reg d ->
      e $ emit 0x81
	$ emit_operand 5 (operand_reg (Reg d))
	$ emit_int s
  | _ -> operand_error2 "sub" s d
;;

let jmp d e = match d with
  | Int t ->
      e $ emit 0xE9
	$ emit_int t
  | _ -> operand_error1 "jmp" d
;;

let jcc c d e = match d with
  | Int t ->
      e $ emit 0x0F
	$ emit (0x80 lor condition_value c)
	$ emit_int t
  | _ -> operand_error1 "jcc" d
;;

(* -------------------------------------------------- *)
(* Wrapper Instructions                               *)
(* -------------------------------------------------- *)

(* Repeat an emitter construct n times.
 * Each application it applied to the resulting emitter of the
 * previous. *)
let rec repeat n f e = match n with
  | 0 -> e
  | n -> repeat (n-1) f (f e)
;; 

(* Repeat an emitter construct n times while passing the current
   decrement as the first argument. *)
let rec repeati n f e = match n with
  | 0 -> e
  | n -> repeati (n-1) f (f (n-1) e)
;;

(* Emit code to push a sequence.
   The sequence is pushed in reverse order,
   ie, starting from the end of the list. *)
let rec push_seq xs e = match xs with
  | [] -> e
  | x :: xs ->
      e $ push_seq xs
	$ push x
;;

(* Emit code to pop a sequence.
   The sequence is popped in the supplied order,
   ie, starting from the beginning of the list. *)
let rec pop_seq xs e = match xs with
  | [] -> e
 (* We ignore integers arguments (can't pop into an integer) *)
  | Int i :: xs ->
      e $ add (Int 4) esp
	$ pop_seq xs
  | x :: xs ->
      e $ pop x
	$ pop_seq xs
;;

(* Push a sequence, emit 'op', and pop back the sequence. *)
let push_pop xs op e =
  e $ push_seq xs
    $ op
    $ pop_seq xs
;;

