#include <stdio.h>
#include <caml/mlvalues.h>
#include <caml/callback.h>
#include <caml/memory.h>
#include <caml/alloc.h>

#include "defines.h"

// save relative instruction pointer
#define SAVE_EIP()                              \
    int eip;                                    \
    asm("movl 4(%%ebp), %0" : "=r"(eip));       \
    eip -= current_code;

// restore relative instruction pointer
#define RESTORE_EIP()                           \
    eip += current_code;                        \
    asm("movl %0, 4(%%ebp)" : : "r" (eip));

// OCaml callbacks
static value* ocaml_identity;
static value* ocaml_mkValue;

// current code block
static value current_code;

// C statics and callbacks
static struct {
    value  res;
    value  tmp;
    value  glo;
    value  lib;
    value  atom;
    void * modify;
    void * mkValue;
    void * print;
    void * slowMkTuple;
    void * fastMkTuple;
} statics;


// C callback to create a new DS value
// uses the OCaml callback to do the work
value mkValue(int type, int data, value env, value vec) {
    CAMLparam2(env, vec);
    SAVE_EIP();
 
    value args[4];
    args[0] = Val_int(type);
    args[1] = Val_int(data);
    args[2] = env;
    args[3] = vec;

    value result = caml_callbackN(*ocaml_mkValue, 4, args);

    RESTORE_EIP();
    CAMLreturn(result);
}

// C callback to create a new tuple
// Precondition: 'size' must be larger then zero
value slowMkTuple(int size, value env, value vec) {
    CAMLparam2(env, vec);
    SAVE_EIP();

    value result = caml_alloc_shr(size, 0);
    while (size) caml_initialize(&Field(result, --size), VAL_VOID);

    RESTORE_EIP();
    CAMLreturn(result);
}

// C callback to create a new tuple
// Precondition: 'size' must be larger then zero and less or equal to
// 'Max_young_wosize' = 256 
// Note: The entries in the tuple contains
// garbage and must be overwritten.
value fastMkTuple (int size, value env, value vec) {
    CAMLparam2(env, vec);
    SAVE_EIP();

    value result = caml_alloc_small(size, 0);

    RESTORE_EIP();
    CAMLreturn(result);
}

// Debug helper to print something from compiled assembly code.
// Usage:
//   e $ push <reg>
//     $ genCall print
//     $ pop  <reg>
void print(int arg) {
    printf("printing: %u\n", (int)arg);
    return;
}

// Garbage collection safe modification (currently unused)
void mymodify(value *fp, value v) {
    // from memory.c: [caml_modify] never calls the GC.
    // so we do not need to secure our variables.
    caml_modify(fp, v);
}

// Initialize the dynamic code environment.
// The OCaml callbacks must be registered in OCaml before invoking
// this function.
CAMLprim value
init (value res, value tmp, value glo, value lib) {
    ocaml_mkValue = (value*)caml_named_value("mkValue");
    ocaml_identity = (value*)caml_named_value("identity");
    statics.modify = &mymodify;
    statics.mkValue = &mkValue;
    statics.slowMkTuple = &slowMkTuple;
    statics.print = &print;
    statics.fastMkTuple = &fastMkTuple;

    // register statics and current_code with the ocaml GC
    caml_register_global_root(&statics.res);
    caml_register_global_root(&statics.tmp);
    caml_register_global_root(&statics.glo);
    caml_register_global_root(&statics.lib);
    caml_register_global_root(&current_code);

    // initialize statics 
    statics.atom = current_code = Atom(0);
    statics.res = res;
    statics.tmp = tmp;
    statics.glo = glo;
    statics.lib = lib;

    return Val_int(&statics);
}

// Execute some dynamically compiled code.
CAMLprim void
exec (value code, value pc, value vec, value env) {
    // Set the currently executing code so we can see if the GC moves
    // it during a callback.
    current_code = code;

    // Set up the special registers (esi and edi).
    // Must be in sync with codegen.cpp.ml
    asm("" : : "S"(vec), "D"(env));

    // Execute the compiled code at offset 'pc'.
    goto *(&((char*)code)[Int_val(pc)]);
}
