open Arg;;
open Vml;;
open Utils;;
open Gc;;

type interp = Bits;;

let use_file = ref None;;
let use_interp = ref Bits;;
let show_result = ref true;;
let show_time = ref false;;
let show_stats = ref false;;

let spec =
  [ (* We only have one interpreter backend
    ("-bits", Unit (fun () -> use_interp := Bits),
     "       -- Run the bitstring interpreter (default)")
  ; *)
    ("-noresult", Unit (fun () -> show_result := false),
     "   -- Omit printing the result")
  ; ("-jit", Unit (fun () -> InterpretBits.enable_jit := true),
     "        -- Enable Just-in-time compilation")
  ; ("-minor", Int (fun k -> Gc.set { (Gc.get()) with Gc.minor_heap_size = k * 256 }),
     "size  -- Size of minor heap in kilobytes (default 128k)")
#ifndef RELEASE
  ; ("-time", Unit (fun () -> show_time := true),
     "       -- Output timing information")
  ; ("-trace",  Int (fun level -> InterpretBits.trace_level := level),
     "level -- Show VM tracing information.")
  ; ("-stats", Unit (fun () -> InterpretBits.jit_stats := true; show_stats := true),
     "      -- Show statistical information")
#endif
  ];;

let msg = "The Optimizing Virtual Machine for DAIMI Scheme
Usage: [options] dsb file
Options:"
;;

let main () =
  parse spec (fun s -> use_file := Some s) msg;
  match !use_file with
    | None -> usage spec msg
    | Some f ->
#ifndef RELEASE
	let t1_parse = if !show_time then gettime () else 0.0 in
#endif
	let p = VMLParser.from_file f in
#ifndef RELEASE
	let t2_eval  = if !show_time then gettime () else 0.0 in
#endif
	let r = match !use_interp with
	  | Bits -> InterpretBits.interpret p in
#ifndef RELEASE
	let t3_stop  = if !show_time then gettime () else 0.0 in
#endif
	  (if !show_result 
	   then fprintf std_formatter "%a\n" Values.print r);
#ifndef RELEASE
	  (if !show_time
	   then fprintf std_formatter
	     "Program ran for %.3f seconds.
Parsing used %.3f seconds.
Evaluation used %.3f seconds.\n"
	     (t3_stop -. t1_parse)
	     (t2_eval -. t1_parse)
	     (t3_stop -. t2_eval));
	  (if !show_stats
	   then let stat = Gc.stat () in
	     fprintf std_formatter
	     "Memory Statistics
 Minor heap size:   %d k
 Major heap size:   %d k
 Major live memory: %d k
 Max   live memory: %d k
 Minor allocations: %0.f k
 Major allocations: %0.f k
 Total allocations: %0.f k
 Minor collections: %d
 Major collections: %d
 Compations:        %d\n"
	       ((Gc.get ()).minor_heap_size / 256)
	       (stat.heap_words / 256)
	       (stat.live_words / 256)
	       (stat.top_heap_words / 256)
	       (stat.minor_words /. 256.0)
	       (stat.major_words /. 256.0)
	       ((stat.minor_words +. stat.major_words -. stat.promoted_words) /. 256.0)
	       stat.minor_collections
	       stat.major_collections
	       stat.compactions)
#endif
;;

main ()
